#!/usr/bin/env python3
# -*- coding: utf-8 -*-
import hashlib

try:
    from poppy.core.db.connector import Connector
    from poppy.core.generic.manager import MultipleClassManager
    from poppy.core.generic.metaclasses import SingletonManager
    from poppy.core.generic.signals import Signal
    from poppy.core.logger import logger
    from poppy.core.properties import Properties
#    from poppy.core.generic.cache import CachedProperty
except ImportError:
    print('POPPy framework is not installed!')

from roc.rpl.constants import VALID_PACKET, INVALID_PACKET_HEADER, INVALID_PACKET_DATA
from roc.rpl.packet_parser.base_packet import BasePacket
from roc.rpl.packet_parser.palisade import palisade_metadata
from roc.rpl.time import Time

__all__ = ['PacketParser']


class PacketParser(object):
    """
    A class to make the parsing of input RPW packets.
    """

    # signal to indicate that a new packet class has been created and that
    # special packets needs to be registered again
    reseted = Signal()

    def __init__(self,
                 idb_version=None,
                 idb_source='MIB',
                 time=Time()):
        """
        Initialize the PacketParser library by loading some information from the
        database.
        """

        # Initialize idb manager
        self.idb_manager = None

        self.idb = (idb_source, idb_version)

        # create the IDB manager
        self.create_idb_manager()

        # signal for errors
        self.extract_error = Signal()

        # Get Time instance
        self.time = time

        palisade_version = (self.idb_version
                            if self.idb_source == 'PALISADE' else None)
        self.palisade_metadata = palisade_metadata(
            palisade_version=palisade_version)

        # List of identified/parsed packets
        self._identified_packets = []
        self._parsed_packets = []

    @property
    def idb_version(self):
        return self._idb_version

    @property
    def idb_source(self):
        return self._idb_source

    @property
    def idb(self):
        return self.idb_manager[(self._idb_source, self._idb_version)]

    @idb.setter
    def idb(self, source_version_tuple):
        # set the idb source/version
        self._idb_source, self._idb_version = source_version_tuple

    @property
    def identified_packets(self):
        return self._identified_packets

    @identified_packets.setter
    def identified_packets(self, values):
        self._identified_packets = values

    @property
    def parsed_packets(self):
        return self._parsed_packets

    @parsed_packets.setter
    def parsed_packets(self, values):
        self._parsed_packets = values

    def reset(self):
        """
        Defines a packet class with its own manager for this instance of the
        PacketParser object.
        """
        self.packet = self._create_packet_class()

        # send signal that a new packet class has been reseted and that some
        # packets may want to register
        self.reseted(self.packet)

        self._identified_packets = []

        self._parsed_packets = []

    @Connector.if_connected('MAIN-DB')
    def create_idb_manager(self):
        """
        Create the IDB manager for easily switching between versions inside the
        PacketParser object.
        """
        # get the manager
        self.idb_manager = Connector.manager['MAIN-DB'].idb_manager

    def add_packet(self, current_packet):
        """
        Add input packet to the Packet class instance of PacketParser
        (Required before parsing data of the packet)

        :param current_packet:
        :return: True if packet has been added, False otherwise
        """

        is_added = False
        try:
            if current_packet.get('status') == VALID_PACKET:
                # Get/create new packet instance (one per palisade_id)
                packet = self.packet(current_packet['palisade_id'])
                # Force srdb_id setting (since compressed packets already
                # initialized has srdb_id = None)
                packet.srdb_id = current_packet['srdb_id']
                # Pass idb_source/idb_version into packet (required for L0
                # writing)
                packet.idb_source = current_packet['idb_source']
                packet.idb_version = current_packet['idb_version']
                packet.category = current_packet['category']
                packet.is_valid = True

                # add the packet
                packet.add()

                is_added = True
            elif current_packet.get('status') == INVALID_PACKET_HEADER:
                logger.warning(f'Packet {current_packet} has an invalid header!')
            elif current_packet.get('status') == INVALID_PACKET_DATA:
                logger.warning(f'Packet {current_packet} has invalid data!')
            else:
                logger.warning(f'Packet {current_packet} is not valid!')
        except Exception:
            logger.exception(f'Packet {current_packet} cannot be added!')

        return is_added

    def identify_packets(self, packet_list,
                         start_time=None,
                         end_time=None,
                         valid_only=True,
                         ):
        """
        Given a list of packets, analyze the binary content
        to identify the packets and allocate space for data.

        :param packet_list: List of
        :param start_time: Only keep packet(s) created/executed after start_time
        :param end_time: Only keep packet(s) created/executed before end_time
        :param valid_only: If True only returns valid packets
        :return: tuple containing list
        """

        # Import identify_packets Cython method
        from roc.rpl.packet_parser.parser.identify_packets import identify_packets

        identified_packets = identify_packets(self, packet_list,
                                              start_time=start_time,
                                              end_time=end_time)

        if valid_only:
            self.identified_packets = PacketParser.packet_status(identified_packets, status=VALID_PACKET)
        else:
            self.identified_packets = identified_packets

        return self.identified_packets

    def parse_packets(self, packet_list,
                      start_time=None,
                      end_time=None,
                      valid_only=True):
        """
        Parse input RPW TM/TC packets and store their information
        in the packet structure.

        :param packet_list: List of raw binary data of the packets
        :param start_time: Only keep packet(s) created/executed after start_time
        :param end_time: Only keep packet(s) created/executed before end_time
        :param valid_only: If True only return valid packets
        :return: Store results in the PacketParser object
        """
        from roc.rpl.packet_parser.parser.parse_packets import parse_packets

        parsed_packets = parse_packets(self, packet_list,
                                       start_time=start_time,
                                       end_time=end_time)

        if valid_only:
            self.parsed_packets = self.packet_status(parsed_packets, status=VALID_PACKET)
        else:
            self.parsed_packets = parsed_packets

        return self.parsed_packets

    def _create_packet_class(meta_data):
        class Packet(
            BasePacket,
            metaclass=SingletonManager,
            manager=MultipleClassManager,
        ):
            """
            Create a class for managing packets to parse
            their data, with flexibility to take into account the particularity
            of the TNR/HFR packets with their micro-commands structure.
            """
            # variable for meta information
            meta = Properties()

            def __init__(self, name,
                         srdb_id=None,
                         category=None,
                         idb_version=None,
                         idb_source=None):

                # store the name of the packet (PALISADE_ID)
                self.name = name

                # Store idb_version and idb_source of the packet
                self.idb_version = idb_version
                self.idb_source = idb_source

                # Store SRDB_ID of the packet
                if not srdb_id:
                    srdb_id = meta_data.palisade_metadata[name]['srdb_id']
                self.srdb_id = srdb_id

                # Store packet category
                if not category:
                    category = meta_data.palisade_metadata[name]['packet_category']
                self.category = category

                # init a counter for storing the presence of a packet or not
                self.counter = 0
                self.internal_counter = 0

                # an attribute for storing the data
                self.data = None

                logger.debug(f'Creating packet {self.name}...')

                # store the parameters associated to this packet
                self.parameters = meta_data.idb.packets[self.name].parameters

                # if there are no parameters, we store just headers and skip
                # data extraction
                self.skip_data = False
                if len(self.parameters) <= 0:
                    self.skip_data = True

                # store also the packet size. This is the byte size without
                # accounting for block repetitions. You can consider that it is
                # the size of the packet with only one occurrence of each block
                self.byte_size = meta_data.idb.packets[self.name].byte_size

                # by default, it is not a compressed packet
                self.is_compressed = False

                # Valid packet flag
                self.is_valid = True

        return Packet

    @staticmethod
    def packet_status(packet_list, status=VALID_PACKET, invert=False):
        """
        Return list of packets with a given status.

        :param packet_list: List of input packets to filter
        :param status: status of packets to be returned (VALID_PACKET by default)
        :param invert: if True apply opposite status filtering
        :return: list of packets with given status
        """
        if invert:
            return [current_packet for current_packet in packet_list
                    if current_packet.get('status') != status]
        else:
            return [current_packet for current_packet in packet_list
                    if current_packet.get('status') == status]

    @staticmethod
    def hex_to_bytes(string):
        """
        Convert an hexadecimal string into a byte array.
        """
        # transform to an array of bytes
        return bytearray.fromhex(string)

    @staticmethod
    def get_packet_sha(packet_data):
        """
        Compute the SHA256 of the input packet.
        TM sha is computed from binary
        TC sha is computed from binary and execution UTC time

        :param packet_data: dictionary containing data of the input packet
        :return: string containing SHA (hexdigest)
        """
        sha = None
        binary = packet_data.get('binary', None)
        type = packet_data.get('type', None)
        if binary and type:
            if type == 'TC':
                utc_time = packet_data.get('utc_time')
                srdb_id = packet_data.get('srdb_id')
                ack_acc_state = packet_data.get('ack_acc_state')
                ack_exe_state = packet_data.get('ack_exe_state')
                srdb_id = packet_data.get('srdb_id')

                raw_sha = hashlib.sha256()
                raw_sha.update(binary.encode('utf-8'))
                raw_sha.update(srdb_id.encode('utf-8'))
                raw_sha.update(ack_acc_state.encode('utf-8'))
                raw_sha.update(ack_exe_state.encode('utf-8'))
                raw_sha.update(utc_time.isoformat().encode('utf-8'))
                sha = str(raw_sha.hexdigest())
            elif type == 'TM':
                raw_sha = hashlib.sha256()
                raw_sha.update(binary.encode('utf-8'))
                sha = str(raw_sha.hexdigest())
            else:
                logger.error(f'Unknown packet type {type} ({packet_data})')
        elif binary:
            # If only binary is known, use it for computing SHA256
            raw_sha = hashlib.sha256()
            raw_sha.update(binary.encode('utf-8'))
            sha = str(raw_sha.hexdigest())
        else:
            logger.error(f'Unknown packet type {type} ({packet_data})')

        return sha

    def get_binary(packet_data):

        binary = None
        if 'binary' in packet_data:
            binary = packet_data['binary']
        elif 'Packet' in packet_data:
            binary = packet_data['Packet']
        elif 'RawBodyData' in packet_data:
            binary = packet_data['RawBodyData']
        else:
            logger.warning('No "binary" keyword in the input packet data')

        return binary



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

from operator import itemgetter

import numpy as np

try:
    from poppy.core.logger import logger
except:
    print('POPPy framework is not installed!')

from roc.rpl.constants import INVALID_PACKET_HEADER, \
    RAW_DATA_DEF, TM_RAW_DATA_DEF, TC_RAW_DATA_DEF

__all__ = ['RawData']


class RawData:

    def __init__(self,
                 packet_list=[],
                 idb_version=None,
                 idb_source=None):

        self.reset()
        self.packet_list = packet_list
        self.idb_version = idb_version
        self.idb_source = idb_source

    def reset(self):
        """Reset RawData instance."""

        self._packet_list = []
        self._invalid_packet_list = []

        # self for PacketParser and idb class
        self._packet_parser = None
        self._idb_version = None
        self._idb_source = None

    @property
    def packet_list(self):
        """
        Get list of packets.

        :return: list of packets
        """
        return self._packet_list

    @packet_list.setter
    def packet_list(self, value):
        """
        Set list of packets.

        :param value: new list of packets
        :return: None
        """
        self._packet_list = value

    @property
    def invalid_packet_list(self):
        """
        Get list of invalid packets

        :return: list of invalid packets
        """
        return self._invalid_packet_list

    @invalid_packet_list.setter
    def invalid_packet_list(self, value):
        """
        Set list of invalid packets

        :param value: new list of invalid packets
        :return: None
        """
        self._invalid_packet_list = value

    @property
    def idb_version(self):
        return self._idb_version

    @idb_version.setter
    def idb_version(self, value):
        if (self._packet_parser is not None and
            self._packet_parser.idb_version != value):
            logger.warning(f'idb_version already defined with '
                         f'value {self._packet_parser.idb_version} in RawData.rpl!')
        else:
            self._idb_version = value

    @property
    def idb_source(self):
        return self._idb_source

    @idb_source.setter
    def idb_source(self, value):
        if (self._packet_parser is not None and
            self._packet_parser.idb_source != value):
            logger.warning(f'idb_source already defined with '
                         f'value {self._packet_parser.idb_source} in RawData.rpl!')
        else:
            self._idb_source = value

    @property
    def packet_parser(self):
        return self._packet_parser

    @packet_parser.setter
    def packet_parser(self, packet_parser_instance):

        # Set PacketParser instance of RawData
        # (Only possible if a PacketParser instance is not already defined with different
        # Packet class instance)
        if (self._packet_parser is not None and
            hasattr(self._packet_parser, 'packet') and
            self._packet_parser.packet != packet_parser_instance.packet):
            logger.warning(f'RawData.packet_parser already defined: '
                           f'please reset the current RawData instance first!')
        else:
            self._packet_parser = packet_parser_instance

            # Ensure that values of idb_version and idb_source attributes always
            # match with the packet_parser ones
            if self._idb_version and self._idb_version != self._packet_parser.idb_version:
                logger.warning(f'Existing idb_version value "{self._idb_version}" '
                               f'will be replaced by {self._packet_parser.idb_version}')
            self._idb_version = self._packet_parser.idb_version

            if self._idb_source and self._idb_source != self._packet_parser.idb_source:
                logger.warning(f'Existing idb_version value "{self._idb_source}" '
                               f'will be replaced by {self._packet_parser.idb_source}')
            self._idb_source = self._packet_parser.idb_source

    @property
    def binary(self):
        """
        Get binary data from packet list.

        :return: list of binary data
        """
        return [pkt['binary'] for pkt in self._packet_list]

    @property
    def packet_scet_time(self):
        """
        Get packet creation SCET time from packet list.

        :return: list of packet creation SCET time
        """
        return [pkt['scet'] for pkt in self._packet_list]

    @property
    def packet_utc_time(self):
        """
        Get packet creation UTC time from packet list.

        :return: list of packet creation UTC time
        """
        return [pkt['utc_time'] for pkt in self._packet_list]

    @property
    def packet_srdb_id(self):
        return self.packet_name

    @property
    def packet_name(self):
        """
        Get list of packet name (SRDB ID).

        :return: list of packet name (SRDB ID)
        """
        return [pkt['srdb_id'] for pkt in self._packet_list]

    @property
    def packet_palisade_id(self):
        return [pkt['palisade_id'] for pkt in self._packet_list]

    @property
    def packet_unique_id(self):
        return [pkt['unique_id'] for pkt in self._packet_list]

    @property
    def packet_dates(self):
        return [pkt['utc_time'].date() for pkt in self._packet_list]

    def add_packet(self, packet_data):
        """
        Add a packet dictionary to the packet_list

        :param packet_data:
        :return: return True if the packet is added, False otherwise
        """
        if packet_data not in self._packet_list:
            self._packet_list.append(packet_data)
            return True
        else:
            return False

    @staticmethod
    def filter(packet_list,
                  by_srdbid_and_utc_time=None,
                  by_palisadeid_and_utc_time=None,
                  by_utc_start=None,
                  by_utc_end=None,
                  by_srdbid=None,
                  by_palisadeid=None,
                  by_utc_time=None,
                  by_binary=None,
                  by_date=None,
                  by_type=None):
        """
        Filter input packet list by attributes (name,
        descr, time, binary).

        :param packet_list: list of packets to filter
        :param by_srdbid_and_utc_time: tuple containing the SRDB_ID and UTC time of the packets to return
        :param by_palisadeid_and_utc_time: tuple containing the PALISADE_ID and UTC time of the packets to return
        :param by_name: SRDB_ID of the packets to return
        :param by_utc_start: Return packets for which UTC time is greater or equal than by_utc_start
        :param by_utc_end: Return packets for which UTC time is lesser than by_utc_end
        :param by_descr: PALISADE ID of the packets to return
        :param by_utc_time: UTC time of the packets (datetime object) to return
        :param by_binary: raw binary data of the packets (hexa string) to return
        :param by_date: date (datetime.date object) of packets to return
        :param by_type: type (TM or TC) of packets to return
        :return: return list of packets filtered
        """

        filtered_packet_list= []

        # Loop over the packets to look for the requested
        for current_packet in packet_list:

            if not current_packet:
                continue

            if (by_srdbid_and_utc_time and
                by_srdbid_and_utc_time == (current_packet['srdb_id'], current_packet['utc_time']) and
                current_packet not in filtered_packet_list):
                filtered_packet_list.append(current_packet)

            if (by_palisadeid_and_utc_time and
                by_palisadeid_and_utc_time == (current_packet['palisade_id'], current_packet['utc_time']) and
                current_packet not in filtered_packet_list):
                filtered_packet_list.append(current_packet)

            if (by_srdbid and
                by_srdbid == current_packet['srdb_id'] and
                current_packet not in filtered_packet_list):
                filtered_packet_list.append(current_packet)

            if (by_palisadeid and
                by_palisadeid == current_packet['palisade_id'] and
                current_packet not in filtered_packet_list):
                filtered_packet_list.append(current_packet)

            if (by_utc_start and
            by_utc_start <= current_packet['utc_time'] and
                current_packet not in filtered_packet_list):
                filtered_packet_list.append(current_packet)

            if (by_utc_end and
            by_utc_end > current_packet['utc_time'] and
            current_packet not in filtered_packet_list):
                filtered_packet_list.append(current_packet)

            if (by_utc_time and
            by_utc_time == current_packet['utc_time'] and
            current_packet not in filtered_packet_list):
                filtered_packet_list.append(current_packet)

            if (by_binary and
            by_binary == current_packet['binary'] and
            current_packet not in filtered_packet_list):
                filtered_packet_list.append(current_packet)

            if (by_date and
            by_date == current_packet['utc_time'].date() and
            current_packet not in filtered_packet_list):
                filtered_packet_list.append(current_packet)

            if (by_type and
            by_type == current_packet['type'] and
            current_packet not in filtered_packet_list):
                filtered_packet_list.append(current_packet)

        return filtered_packet_list

    @staticmethod
    def check_packet_name(packet_list, expected_names):
        """
        Check input packets against expected list of expected packet names (SRDB_ID)

        :param packet_list: list of input packets to check
        :param expected_names: list of expected packet names
        :return: True if the checking has passed, else False (if wrong input, then return None)
        """

        npkt = len(packet_list)
        nexp = len(expected_names)
        if nexp != npkt:
            logger.error(f'Input "expected_names" '
                         f'must have {npkt} element(s), but {nexp} found!')
            return None

        for i, current_packet in enumerate(packet_list):

            expected_name = expected_names[i]

            packet_name = current_packet['srdb_id']

            try:
                logger.debug('Comparing reconstructed names with given one: '
                             f'{packet_name} ?= {expected_name}')
                assert packet_name == expected_name
            except:
                # If not the same name, then set the status of the packet to invalid
                packet_list[i]['status'] = INVALID_PACKET_HEADER
                logger.warning(f'Packet #{i} has a wrong name: '
                               f'"{packet_name}" found but "{expected_name}" expected ')
                continue

        return packet_list

    @staticmethod
    def by_increasing_time(packet_list, decreasing=False, by_scet=False):
        """
        Sort input packet list by increasing packet time.

        :param deacreasing: if True, then sort by decreasing scet
        :param by_scet: if True, use scet time instead of utc_time
        :return:
        """

        if by_scet == True:
            packet_time = [pkt['scet'] for pkt in packet_list]
        else:
            packet_time = [pkt['utc_time'] for pkt in packet_list]

        if packet_time:
            index_list = sorted(range(len(packet_time)), key=lambda k: packet_time[k])
            if decreasing:
                index_list = reversed(index_list)

            return list(itemgetter(*index_list)(packet_list))
        else:
            return None


    def is_packet(self, packet_id):
        """
        Check if a given packet exists in the current PacketParser object

        :param packet_id: ID of the packet (can be SRBD ID or PALISADE ID)
        :return: True if packet has been found, False otherwise
        """

        # Check if input packet id is a PALISADE or SRBD ID
        if (packet_id.startswith('TM_') or
            packet_id.startswith('TC_')):
            packet_ids = self.packet_palisade_id
        elif (packet_id.startswith('YIW') or
            packet_id.startswith('ZIW')):
            packet_ids = self.packet_name
        else:
            logger.error(f'Unknown input packet id: {packet_id}')
            return False

        return packet_id in packet_ids


    @staticmethod
    def sort_by_utc_time(packet_list,
                         packet_utc_time=None):
        """
        Return input packet_list sorted by increasing utc time.


        :return: sorted packet_list
        """

        if packet_utc_time is None:
            packet_utc_time = [pkt['utc_time'] for pkt in packet_list]

        # get indices of sorted utc_time list
        sorted_indices = np.argsort(packet_utc_time)

        return list(np.array(packet_list)[sorted_indices])

    @staticmethod
    def is_identified(packet_data):
        """
        Check if a packet has been already identified.

        :param packet_data: packet data as a dictionary
        :return: True, if the packet has been already identified, False otherwise
        """

        # Check if generic parameters are present
        for key in RAW_DATA_DEF.keys():
            # Get parameter value from packet_data dictionary
            # If not found, then set to NoneType value by default
            if not packet_data.get(key, None):
                # If at least one of the parameter is not defined (i.e., has NoneType value)
                # then return False
                return False

        # Additional checks
        # srdb id format should be as expected
        if (not packet_data['srdb_id'].startswith('YIW') and not
        packet_data['srdb_id'].startswith('ZIW')):
            return False

        # status should be 'VALID'
        if packet_data['status'] != 'VALID':
            return False

        # Check if generic parameters are present
        if packet_data['type'] == 'TM':
            raw_data_def_dict = TM_RAW_DATA_DEF
        elif packet_data['type'] == 'TC':
            raw_data_def_dict = TC_RAW_DATA_DEF
        else:
            # if not expected packet type, then return false
            return False

        # Loop on raw_data_def_dict default keywords
        for key in raw_data_def_dict.keys():
            # Get parameter value from packet_data dictionary
            # If not found, then set the NoneType value by default
            if not packet_data.get(key, None):
                # If at least one of the parameter is not defined (i.e., has NoneType value)
                # then return False
                return False

        return True

    @staticmethod
    def init_packet_data(packet_data):
        """
        Initialize content of packet_data dictionary

        :param packet_data: dictionary with packet data
        :return: packet_data with default values
        """

        for key, val in RAW_DATA_DEF.items():
            # if keyword (key) not found in packet_data, then set default value (val)
            packet_data[key] = packet_data.get(key, val)

        return packet_data



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import os
from glob import glob
from datetime import datetime
import re

import numpy as np

from spice_manager import SpiceManager

from poppy.core.logger import logger
from poppy.core.configuration import Configuration

from roc.rpl.constants import TIME_DAILY_STRFORMAT, SPICE_KERNEL_PATTERN, KERNEL_CATEGORY
from roc.rpl.exceptions import SpiceKernelNotValid

__all__ = ['SpiceHarvester']

STRPTIME = '%Y%m%d'


class SpiceHarvester(SpiceManager):
    """
    Class to get SPICE kernels for Solar Orbiter (inherits SpiceManager class)
    """

    def __init__(self, path):

        self.path = path
        self._kernels_time_version = {}
        self._failed = False

        super().__init__(spice_kernels=None,
                        logger=logger)

        # Make sure path in the OS environment variables (for kernel loading in mk)
        os.environ['SPICE_KERNEL_PATH'] = path

    @property
    def failed(self):
        return self._failed

    @failed.setter
    def failed(self, value):
        if isinstance(value, bool):
            self._failed = value

    @staticmethod
    def spice_kernel_path():

        # Get pipeline env. variable from configuration file
        try:
            env_vars = Configuration.manager['pipeline']['environment']
        except:
            logger.exception('No "environment" field defined in the pipeline configuration file')
            env_vars = None

        # If SPICE_KERNEL_PATH is provided, then load it
        if env_vars and 'SPICE_KERNEL_PATH' in env_vars:
            if env_vars['SPICE_KERNEL_PATH'].startswith('$'):
                spice_kernel_path = os.environ['SPICE_KERNEL_PATH']
            else:
                spice_kernel_path = env_vars['SPICE_KERNEL_PATH']
            logger.debug(f'SPICE_KERNEL_PATH = {spice_kernel_path} loaded from config. file')
        elif 'SPICE_KERNEL_PATH' in os.environ:
            spice_kernel_path = os.environ['SPICE_KERNEL_PATH']
            logger.debug(f'SPICE_KERNEL_PATH = {spice_kernel_path} loaded from OS environment')
        else:
            logger.info('No entry found for SPICE_KERNEL_PATH')
            spice_kernel_path = None

        return spice_kernel_path

    @classmethod
    def load_spice_kernels(cls,
                        path,
                        by_date=None,
                        ignore_type=[],
                        only_mk=False,
                        predictive=True,
                        flown=True,
                        ):
        """
        Get list of spice kernels to load for Solar Orbiter.

        :param path: Path of spice kernel root directory
        :param by_date: Get spice kernels valid for a given date. Input date value must a datetime.date object. If not provided try to load latest kernels.
        :param ignore_type: Filter spice kernels by type. Must be a list of extension without dot (e.g., "spk", "bsp", etc.). If not provided, then take all types.
        :param ignore_file: List of spice kernel file(s) to ignore.
        :param include_file: List of spice kernel file(s) to be included.
        :param only_mk: Load only meta kernels
        :param predictive: If True then load predictive kernels only, otherwise load predictive then as-flown
        :param flown: If True then load as-flown kernels only, otherwise load predictive then as-flown
        :return: list of spice kernels found
        """

        # If path is None, return None
        if not path or not os.path.isdir(path):
            logger.warning(f'Spice kernel directory path is not valid! ({path})')
            return None

        # Initialize harvester
        harvester = cls(path)

        if not by_date:
            logger.debug('Attempting to load latest SPICE kernels.')
            by_date = datetime.today()
            latest = True
        else:
            logger.debug(f'Loading SPICE kernels for {by_date}')
            latest = False

        if only_mk:
            logger.debug('Loading meta kernels "mk" only')
            # If only meta kernels requested, add all other kernel types in the
            # kernel to ignore
            ignore_type = list(SPICE_KERNEL_PATTERN.keys()).copy()
            ignore_type.remove('mk')

        # Set the kernel category (predictive and/or as-flown)
        if predictive and not flown:
            kernel_categ = ['pred']
        elif not predictive and flown:
            kernel_categ = ['flown']
        else:
            kernel_categ = KERNEL_CATEGORY

        # Check for spice kernels in "path"
        if os.path.isdir(path):

            # Walk through each spice kernels types
            # and get list of kernels found
            for current_type, current_pattern in SPICE_KERNEL_PATTERN.items():

                if current_type in ignore_type:
                    logger.debug(f'{current_type} SPICE kernels will be ignored')
                    continue

                # Define kernel sub-folder
                kernel_subdir = os.path.join(path, current_type)
                if not os.path.isdir(kernel_subdir):
                    logger.error(f'{kernel_subdir} does not exit, skip it')
                    continue

                # Initialize kernel list
                kernel_list = []
                if current_type == 'mk':
                    # Load other kernels from the meta kernel directory
                    os.chdir(os.path.join(path, 'mk'))
                    if latest:
                        # Load latest metakernel
                        suffix = 'mk'
                        kernel_list = [
                            os.path.join(kernel_subdir, f'solo_ANC_soc-{current_categ}-{suffix}.tm')
                            for current_categ in kernel_categ]
                    elif by_date:
                        # Load metakernel(s) for a given date
                        kernel_list = SpiceHarvester.load_mk(kernel_subdir, by_date,
                                                             category=kernel_categ)
                else:
                    logger.debug('Only "mk" SPICE kernel processing is supported currently')
                    continue

                if not kernel_list:
                    logger.warning(f'No SPICE kernel found for {current_type}')
                else:
                    # Load kernels
                    harvester.kernels = kernel_list

        else:
            logger.error(f'{path} does not exists!')
            return None

        return harvester

    @staticmethod
    def load_mk(mk_dir_path, by_date, category=KERNEL_CATEGORY):
        """
        Load SPICE metakernel (mk)

        :param mk_dir_path: Path of the mk directory
        :param by_date: date for which kernel must be loaded (take latest close date)
        :param category: list with kernel category ("flown" and/or "pred")
        :return: list of mk found
        """

        # Get file pattern for the mk kernel type
        kernel_pattern = SPICE_KERNEL_PATTERN['mk'][0]

        # Filter kernel by type
        pattern = os.path.join(mk_dir_path, kernel_pattern)
        mk_files = glob(pattern)
        nmk = len(mk_files)
        if nmk == 0:
            logger.warning(f'No "mk" kernel found in {mk_dir_path}!')
            return None
        else:
            logger.debug(f'{nmk} "mk" kernel(s) found in {mk_dir_path}')

        # For each file get the file category version, date and count
        versions = np.empty(nmk, dtype=int)
        dates = np.empty(nmk, dtype=datetime)
        counts = np.empty(nmk, dtype=int)
        for i, current_mk in enumerate(mk_files):

            # Get category, version date and count from filename fields
            fields = SpiceHarvester.extract_kernel_fields('mk', current_mk)

            if fields['category'] not in category:
                logger.info(f'{current_mk} has not the expected category: {fields["category"]}')
            else:
                versions[i] = fields['version']
                dates[i] = fields['date']
                counts[i] = fields['count']

        # Keep only the nearest but older kernel files w.r.t. by_date
        delta_dt = np.abs(dates - by_date)
        where_min_dt = np.where(delta_dt == min(delta_dt) and dates >= by_date)

        # If remaining kernels, keep the indices for max version
        if len(where_min_dt) > 1:
            where_max_version = np.where(versions[where_min_dt] == max(versions[where_min_dt]))

            # If remaining kernels, keep the indices for max count
            if len(where_max_version) > 1:
                where_max_count = np.where(versions[where_min_dt] == max(versions[where_min_dt]))

                idx = where_min_dt[where_max_version[where_max_count]]
            else:
                idx = where_min_dt[where_max_version]
        else:
            idx = where_min_dt

        # Return only slice of list for the resulting indices
        if idx:
            nearest_mk = mk_files[idx]
        else:
            logger.warning(f'Cannot determinate the "mk" kernel(s) to load for {by_date}')
            nearest_mk = []

        return nearest_mk

    @staticmethod
    def extract_kernel_fields(kernel_type, kernel_file):
        """
        Extract the filename fields from an input SPICE kernel file.

        :param kernel_type: Type of SPICE kernel
        :param kernel_file: SPICE kernel file for which must be extracted
        :return: dictionary with extracted fields
        """

        # regex pattern
        regex_field_pattern = SPICE_KERNEL_PATTERN[kernel_type][1]
        regex_field_num = SPICE_KERNEL_PATTERN[kernel_type][2]

        # Initialize output dictionary
        kernel_fields = {}

        # Retrieve date and version from kernel filename
        basename = os.path.basename(kernel_file)
        try:
            re_result = re.search(regex_field_pattern, basename)
            fields = re_result.groups()
            nfields = len(fields)
            # If there is not the expected number of fields then raise an exception
            if nfields != regex_field_num:
                raise SpiceKernelNotValid
        except Exception:
            logger.exception(f'Wrong file format for {kernel_file}')
        else:
            # Extract fields, depending of the type of kernel
            if kernel_type == 'mk':
                # Get pred/flown category from file name (e.g., solo_ANC_soc-flown-mk.tm)
                category = basename.split('_')[2].split('-')[1]
                kernel_fields = {'version': int(fields[0]),
                                  'date': datetime.strptime(fields[1], TIME_DAILY_STRFORMAT),
                                  'count': int(fields[2]),
                                  'starttime': None,
                                  'endtime': None,
                                  'category': category}
            else:
                logger.warning('Filtering other kernel type than "mk" not supported currently')

        return kernel_fields



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
import uuid

from roc.rpl.tasks.tools import get_raw_data

try:
    from poppy.core.logger import logger
    from poppy.core.task import Task
    from poppy.core.target import PyObjectTarget
    from poppy.core.tools.exceptions import MissingInput
    from poppy.core.db.connector import Connector
except:
    print('POPPy framework is not installed!')

from roc.rpl import VALID_PACKET, PIPELINE_DATABASE
from roc.rpl.exceptions import PacketParsingError
from roc.rpl.time import Time
from roc.rpl.packet_parser.packet_parser import PacketParser

__all__ = ['IdentifyPackets']

class IdentifyPackets(Task):
    """
    Task to perform the RPW packet identification.
    """

    plugin_name = 'roc.rpl'
    name = 'identify_packets'

    def add_targets(self):
        self.add_input(identifier='raw_data',
                       value=get_raw_data(self.pipeline),
                       target_class=PyObjectTarget)
        self.add_output(identifier='raw_data',
                        target_class=PyObjectTarget)

    @Connector.if_connected(PIPELINE_DATABASE)
    def setup_inputs(self):
        # Get input target raw_data value
        try:
            self.raw_data = self.inputs['raw_data'].value
        except:
            raise MissingInput('raw_data input is missing')

        # Get/create Time instance for time computation
        self.time_instance = Time()

        # Get IDB inputs
        self.idb_version = self.pipeline.get('idb_version',
                                        default=[self.raw_data['idb_version']])[0]
        self.idb_source = self.pipeline.get('idb_source',
                                       default=[self.raw_data['idb_source']])[0]
        # If idb_version not passed, try to get current IDB version from database
        if self.idb_version is None:
            from roc.idb.tools.db import get_current_idb
            session = Connector.manager[PIPELINE_DATABASE].session
            self.idb_version = get_current_idb(self.idb_source, session)

        # Pass input arguments for the Time instance
        self.time_instance.kernel_date = self.pipeline.get('kernel_date', default=None, args=True)
        self.time_instance.predictive = self.pipeline.get('predictive', default=True, args=True)
        self.time_instance.no_spice = self.pipeline.get('no_spice', default=False, args=True)

        # Initialize packet parser
        self.parser = PacketParser(
                idb_version=self.idb_version,
                idb_source=self.idb_source,
                time=self.time_instance,
            )

        # Get start-time
        self.start_time = self.pipeline.get('start_time', default=[None])[0]

        # Get end-time
        self.end_time = self.pipeline.get('end_time', default=[None])[0]

    def run(self):

        # Define task job ID (long and short)
        self.job_uuid = str(uuid.uuid4())
        self.job_id = f'IdentifyPacket-{self.job_uuid[:8]}'
        logger.debug(f'Task {self.job_id} is starting')
        try:
            self.setup_inputs()
        except:
            logger.exception(
                f'Initializing inputs has failed for {self.job_id}!')
            self.pipeline.exit()
            return

        # connect to add exception when packet analysis is bad
        self.parser.extract_error.connect(self.exception)

        # Analyse input RPW TM/TC packets
        packet_list = self.raw_data['packet_list']
        logger.info(f'Identifying {len(packet_list)} input packets...')
        try:
            analyzed_packet_list = self.parser.identify_packets(
                                            packet_list,
                                            start_time=self.start_time,
                                            end_time=self.end_time,
                                            valid_only=False)
        except:
            raise PacketParsingError('Packet parsing has failed!')
        else:
            self.raw_data['packet_list'] = [current_packet
                                    for current_packet in analyzed_packet_list
                                    if current_packet['status'] == VALID_PACKET]
            self.raw_data['invalid_packet_list'] = [current_packet
                                    for current_packet in analyzed_packet_list
                                    if current_packet['status'] != VALID_PACKET]
            self.raw_data['packet_parser'] = self.parser

        # store the rpl instance into the properties
        self.outputs['raw_data'].value = self.raw_data



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

from roc.idb.constants import IDB_SOURCE

from roc.rpl import SCOS_HEADER_BYTES
from roc.rpl.tasks import IdentifyPackets, ParsePackets
from roc.rpl.tasks.packets_to_json import PacketsToJson

try:
    from poppy.core.command import Command
    from poppy.core.logger import logger
except:
    print('POPPy framework is not installed!')

__all__ = ['RPL']


class RPL(Command):
    """
    A command to do the decommutation of a test log file in the XML format.
    """
    __command__ = 'rpl'
    __command_name__ = 'rpl'
    __parent__ = 'master'
    __parent_arguments__ = ['base']
    __help__ = (
        'Commands relative to the decommutation of packets with help ' +
        'of PacketParser library'
    )
    def add_arguments(self, parser):
        """
        Add input arguments common to all the RPL plugin.

        :param parser: high-level pipeline parser
        :return:
        """

        # specify the IDB version to use
        parser.add_argument(
            '--idb-version',
            help='IDB version to use.',
            default=[None],
            nargs=1,
        )

        # specify the IDB source to use
        parser.add_argument(
            '--idb-source',
            help='IDB source to use: SRDB, PALISADE or MIB.',
            default=[IDB_SOURCE],
            nargs=1,
        )

        # Remove SCOS2000 header in the binary packet
        parser.add_argument(
            '--scos-header',
            nargs=1,
            type=int,
            default=[SCOS_HEADER_BYTES],
            help='Length (in bytes) of SCOS2000 header to be removed'
                 ' from the TM packet in the DDS file.'
                 f' (Default value is {SCOS_HEADER_BYTES} bytes.)',
        )

class PacketsToJsonCommand(Command):
    """
    Command to parse and write RPW packets data into a JSON format file.

    input binary data must be passed as hexadecimal strings.
    IDB must be available in the database.
    """
    __command__ = 'rpl_pkt_to_json'
    __command_name__ = 'pkt_to_json'
    __parent__ = 'rpl'
    __parent_arguments__ = ['base']
    __help__ = """
        Command to parse and save RPW packets data into a JSON format file.
    """

    def add_arguments(self, parser):

        # packet binary data
        parser.add_argument(
            '-b', '--binary',
            help="""
             Packet binary data (hexadecimal).
             """,
            type=str,
            nargs='+',
            required=True,
        )

        # Output file path
        parser.add_argument(
            '-j', '--output-json-file',
            help=f'If passed, save packet data in a output JSON file.',
            type=str,
            nargs=1,
            default=[None],
        )

        # Header only
        parser.add_argument(
            '-H', '--header-only',
            action='store_true',
            default=False,
            help='Return packet header information only'
        )

        # Only return valid packets
        parser.add_argument(
            '-V', '--valid-only',
            action='store_true',
            default=False,
            help='Return valid packets only'
        )

        # Only return valid packets
        parser.add_argument(
            '-P', '--print',
            action='store_true',
            default=False,
            help='Return results in stdin'
        )

        # Overwrite existing file
        parser.add_argument(
            '-O', '--overwrite',
            action='store_true',
            default=False,
            help='Overwrite existing JSON file'
        )

    def setup_tasks(self, pipeline):
        """
        Execute the command.
        """

        if pipeline.args.header_only:
            start = IdentifyPackets()
        else:
            start = ParsePackets()

        pipeline | start | PacketsToJson()

        # define the start points of the pipeline
        pipeline.start = start



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['roc',
 'roc.rpl',
 'roc.rpl.compressed',
 'roc.rpl.packet_parser',
 'roc.rpl.packet_parser.parser',
 'roc.rpl.packet_structure',
 'roc.rpl.rice',
 'roc.rpl.tasks',
 'roc.rpl.tests',
 'roc.rpl.time']

package_data = \
{'': ['*'], 'roc.rpl.tests': ['data/*']}

install_requires = \
['cython>=3,<4',
 'maser-tools',
 'numpy!=1.19.5',
 'poppy-core>=0.9.4',
 'poppy-pop>=0.7.5',
 'roc-idb>=1.0,<2.0',
 'spice_manager']

setup_kwargs = {
    'name': 'roc-rpl',
    'version': '1.6.0',
    'description': 'RPW Packet parsing Library (RPL): a plugin for the RPW TM/TC packet analysis',
    'long_description': 'RPL PLUGIN README\n=================\n\n[![pipeline status](https://gitlab.obspm.fr/ROC/Pipelines/Plugins/RPL/badges/develop/pipeline.svg)](https://gitlab.obspm.fr/ROC/Pipelines/Plugins/RPL/pipelines)\n\nThis directory contains the source files of the RPW Packet parsing Library (RPL), a plugin of the ROC pipelines dedicated to parse the RPW telemetry/command packets.\n\nRPL has been developed with the [POPPY framework](https://poppy-framework.readthedocs.io/en/latest/).\n\n## Quickstart\n\n### Installation with pip\n\nTo install the plugin using pip:\n\n```\npip install roc-rpl\n```\n\nNOTES:\n\n    - It is also possible to install plugin from gitlab: `pip install roc-rpl --extra-index-url https://__token__:<your_personal_token>@gitlab.obspm.fr/api/v4/projects/2052/packages/pypi/simple --trusted-host gitlab.obspm.fr`. A personal access token is required to reach the package registry in the ROC Gitlab server.\n\n### Installation from the repository\n\nFirst, retrieve the `RPL` repository from the ROC gitlab server:\n\n```\ngit clone https://gitlab.obspm.fr/ROC/Pipelines/Plugins/RPL.git\n```\n\nThen, install the package (here using (poetry)[https://python-poetry.org/]):\n\n```\npoetry install --extras "poppy"\n```\n\nNOTES:\n\n    - It is also possible to clone the repositiory using SSH\n    - To install poetry: `pip install poetry`\n    - Default branch is `develop`\n\n## Usage\n\nThe roc-rpl plugin is designed to be run in a POPPy-built pipeline.\nNevertheless, it is still possible to import some classes and methods in Python files.\n\nFor instance, to test that the installation has ended correctly, run:\n\n```\npython -c "from roc.rpl import packet_structure"\n```\n\nNo message should be returned if the import works well.\n\n## Authors\n\n* Xavier BONNIN xavier.bonnin@obspm.fr (maintainer)\n* Manuel DUARTE manuel.duarte@obspm.fr (author)\n* Sonny LION sonny.lion@obspm.fr (author)\n\nLicense\n-------\n\nThis project is licensed under CeCILL-C.\n\nAcknowledgments\n---------------\n\n* Solar Orbiter / RPW Operation Centre (ROC) team\n',
    'author': 'Xavier Bonnin',
    'author_email': 'xavier.bonnin@obspm.fr',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://gitlab.obspm.fr/ROC/Pipelines/Plugins/RPL',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<4',
}
from build_cython import *
build(setup_kwargs)

setup(**setup_kwargs)



from roc.rpl.rice cimport crice
import logging

from libc.stdint cimport uint8_t
from libc.stdint cimport uint16_t
from libc.stdint cimport uint32_t
from libc.stdint cimport uint64_t
from libc.stdint cimport int32_t
from cpython.mem cimport PyMem_Malloc
from cpython.mem cimport PyMem_Free
from roc.rpl.packet_structure.data cimport Data

__all__ = ["compress", "uncompress"]

logger = logging.getLogger(__name__)


class RiceException(Exception):
    """
    Exceptions for the RICE library.
    """



cdef class Compressor:
    cdef uint32_t* int_buffer
    cdef uint16_t* short_buffer
    cdef uint8_t* byte_buffer
    cdef uint32_t _buffer_size

    def __cinit__(Compressor self, uint32_t buffer_size=0x4000):
        self.buffer_size = buffer_size

    cpdef Data uncompress(
        Compressor self,
        Data data,
        uint32_t output_size,
    ):
        cdef:
            int32_t decompressedSize
            int32_t postProcessedSize
            int32_t buffer_int_size

        # check valid sizes
        if output_size > self._buffer_size:
            raise RiceException("Internal buffer too small for input")

        # decompress the data and get the size
        decompressedSize = crice.uncompress(
            # self.byte_buffer,
            data.data,
            data.size,
            self.int_buffer,
            output_size // 2,
        )
        if decompressedSize == -1:
            raise RiceException("Error uncompressing data")

        # post process the data
        postProcessedSize = crice.postprocessor(
            self.int_buffer,
            decompressedSize,
            self.short_buffer,
            crice.ESTIMATE_1D_H,
        )
        if postProcessedSize == -1:
            raise RiceException("Error post processing decompressed data")

        # compute byte size
        postProcessedSize = postProcessedSize * sizeof(uint16_t)

        return Data(
            (<uint8_t*>self.short_buffer)[:postProcessedSize],
            postProcessedSize,
        )

    cpdef Data compress(
        Compressor self,
        Data data,
    ):
        cdef:
            uint8_t* compressed
            int32_t result
            Py_ssize_t i
            int32_t preProcessedSize
            uint16_t* tmp

        # check sizes are correct with the chosen buffers
        if data.size // 2 > self._buffer_size:
            raise RiceException("Internal buffers too small for compression")

        # pre-processed data generation
        preProcessedSize = crice.preprocess(
            <uint16_t*>data.data,
            data.size // 2,
            self.int_buffer,
            crice.ESTIMATE_1D_H,
        )
        if preProcessedSize == -1:
            raise RiceException("Error pre-processing data before compression.")

        # compress the data
        result = crice.compress(
            self.int_buffer,
            preProcessedSize,
            self.byte_buffer,
            self._buffer_size,
        )
        if result == -1:
            raise RiceException("Error occurred while compressing data")
        if <uint32_t>result > self._buffer_size:
            raise RiceException(
                "Byte buffer too small for compression.\n" +
                "Buffer size: {0} bytes, compressed size: {1} bytes".format(
                    self._buffer_size,
                    result,
                )
            )

        # copy the result
        return Data(self.byte_buffer[:result], result)

    property buffer_size:
        def __get__(self):
            return self._buffer_size

        def __set__(self, buffer_size):
            cdef:
                Py_ssize_t i
            # store the new buffer size
            self._buffer_size = buffer_size

            # if memory already allocated, deallocate
            if self.int_buffer != NULL:
                PyMem_Free(self.int_buffer)
            if self.short_buffer != NULL:
                PyMem_Free(self.short_buffer)
            if self.byte_buffer != NULL:
                PyMem_Free(self.byte_buffer)

            # allocate buffers with the new size
            self.int_buffer = <uint32_t*> PyMem_Malloc(
                buffer_size * sizeof(uint32_t)
            )
            self.short_buffer = <uint16_t*> PyMem_Malloc(
                buffer_size * sizeof(uint16_t)
            )
            self.byte_buffer = <uint8_t*> PyMem_Malloc(
                buffer_size * sizeof(uint8_t)
            )
            for i in range(buffer_size):
                self.byte_buffer[i] = 0
                self.short_buffer[i] = 0
                self.int_buffer[i] = 0

        def __del__(self):
            # if memory already allocated, deallocate
            if self.int_buffer != NULL:
                PyMem_Free(self.int_buffer)
            if self.short_buffer != NULL:
                PyMem_Free(self.short_buffer)
            if self.byte_buffer != NULL:
                PyMem_Free(self.byte_buffer)

    def __dealloc__(self):
        del self.buffer_size
